(* $Id: godi_match.ml 1093 2012-08-11 22:51:55Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_match_lex

(**********************************************************************)
(* GLOBBING, FNMATCH                                                  *)
(**********************************************************************)

type glob_expr_atom =
    G_literal of string
      (* The string must not be empty. Two adjecent G_literal atoms are
       * not allowed.
       * The backslash is not an escape character.
       *)
  | G_star
  | G_qmark
  | G_bracket of (bool * set)   (* bool: whether negated *)
  | G_brace of glob_expr list

and set =
    (int * int) list
  (* Sets must not be empty. For every pair (x,y): x<=y *)


and glob_expr =
    Glob of glob_expr_atom list
;;

exception Bad_glob_expr of string

exception Bad_regular_expr of string

exception Unsupported_expr of string


let literal_glob_expr s =
  Glob [ G_literal s ]


let parse_glob_expr
      ?(enable_star = true)
      ?(enable_qmark = true)
      ?(enable_brackets = true)
      ?(enable_braces = true)
      ?(enable_escape = true)
      s =
  
  let feat =
    { enable_star = enable_star;
      enable_qmark = enable_qmark;
      enable_brackets = enable_brackets;
      enable_braces = enable_braces;
      enable_escape = enable_escape;
      escaped = false;
    } in

  let rec collect_until lexbuf =
    let tok = glob_expr feat lexbuf in
    if tok = Glob_end then
      []
    else
      tok :: (collect_until lexbuf)
  in

  let rec process_brace_list current list =
    match list with
	Brace_literal s :: list' ->
	  let gl = collect_until (Lexing.from_string s) in
	  process_brace_list (current @ gl) list'
      | Brace_braces l :: list' ->
	  process_brace_list (current @ [Glob_braces l]) list'
      | Brace_comma :: list' ->
	  let ge = process_glob_list [] current in
	  Glob ge :: process_brace_list [] list'
      | Brace_end :: _ ->
	  assert false
      | [] ->
	  let ge = process_glob_list [] current in
	  [ Glob ge ]

  and process_glob_list acc list =
    match list with
	Glob_star :: list' ->
	  ( match acc with
		G_star :: acc' ->
		  (* Ignore the second star! *)
		  process_glob_list acc list'
	      | _ ->
		  process_glob_list (G_star :: acc) list'
	  )
      | Glob_qmark :: list' ->
	  process_glob_list (G_qmark :: acc) list'
      | Glob_brackets (neg,btoks) :: list' ->
	  let set =
	    List.map
	      (function
		   Bracket_char c ->
		     (Char.code c, Char.code c)
		 | Bracket_range (c1,c2) ->
		     (Char.code c1, Char.code c2)
		 | Bracket_end ->
		     assert false
	      )
	      btoks in
	  process_glob_list (G_bracket(neg,set) :: acc) list'
      | Glob_braces btoks :: list' ->
	  let alts = process_brace_list [] btoks in
	  process_glob_list (G_brace alts :: acc) list'
      | Glob_literal s :: list' ->
	  if s <> "" then
	    ( match acc with
		  G_literal s' :: acc' ->
		    process_glob_list (G_literal(s' ^ s) :: acc') list'
		| _ ->
		    process_glob_list (G_literal s :: acc) list'
	    )
	  else
	    process_glob_list acc list'
      | Glob_end :: _ ->
	  assert false
      | [] ->
	  List.rev acc
  in

  try
    let glob_list = 
      collect_until (Lexing.from_string s) in
    
    let glob_expr =
      process_glob_list [] glob_list in

    Glob glob_expr
  with
      Bracket_Unsupported ->
	raise (Unsupported_expr s)
    | Lexing_Error ->
	raise (Bad_glob_expr s)
;;


let esc_re = Str.regexp "[][*?{},\\]";;

let esc_subst s =
  "\\" ^ Str.matched_string s
;;


let rec mem_set code set =
  match set with
      [] -> false
    | (x,y) :: set' ->
	(x <= y && code >= x && code <= y) || mem_set code set'
;;


let size_set set =
  (* Please don't tell anybody... *)
  let n = ref 0 in
  for k = 0 to 255 do
    if mem_set k set then incr n
  done;
  !n
;;


let rec exclude_set codes set =
  match set with
      [] -> []
    | (x,y) :: set' ->
	let x' = if List.mem x codes then x+1 else x in
	let y' = if List.mem y codes then y-1 else y in
	if x = x' && y = y' && x <= y then
	  (x,y) :: exclude_set codes set'
	else if x' <= y' then 
	  exclude_set codes ( (x',y') :: set')
	else
	  exclude_set codes set'
;;


let print_set buf neg_char negated set =
  (* Always produce a portable expression: *)

  (* Check for special characters: *)
  let want_minus    = mem_set (Char.code '-') set in
  let want_rbracket = mem_set (Char.code ']') set in
  let want_circum   = mem_set (Char.code '^') set in
  let want_exclam   = mem_set (Char.code '!') set in
  let size = size_set set in
  
  (* Check for very special sets: *)
  if not negated && want_circum && size = 1 then
    Buffer.add_string buf "^"  (* "[^]" would not be portable enough *)
  else if not negated && want_exclam && size = 1 then
    Buffer.add_string buf "!"  (* "[!]" would not be portable enough *)
  else if not negated && want_circum && want_exclam && size = 2 then
    failwith "print_glob_expr"
      (* There is no portable representation *)
  else (
    (* First create a set expression where the special characters
     * '-', ']', '^', and '!' do not occur literally.
     *)
    let empty = ref true in
    let buf' = Buffer.create 200 in
    let set' = 
      exclude_set (List.map Char.code ['-'; ']'; '^'; '!']) set in
    List.iter
      (fun (x0,x1) ->
	 if x0 = x1 then (
	   Buffer.add_char buf' (Char.chr x0);
	   empty := false;
	 ) 
	 else if x0 <= x1 then (
	   Buffer.add_char buf' (Char.chr x0);
	   Buffer.add_char buf' '-';
	   Buffer.add_char buf' (Char.chr x1);
	   empty := false;
	 )
      )
      set';
    (* Check which of the special characters are already covered
     * by ranges:
     *)
    let done_minus    = mem_set (Char.code '-') set' in
    let done_rbracket = mem_set (Char.code ']') set' in
    let done_circum   = mem_set (Char.code '^') set' in
    let done_exclam   = mem_set (Char.code '!') set' in
    (* Begin with printing *)
    Buffer.add_string
      buf 
      (if negated then "[" ^ String.make 1 neg_char else "[");
    (* ']' must always be the first character of the set: *)
    if want_rbracket && not done_rbracket then (
      Buffer.add_string buf "]";
      empty := false;
    );
    Buffer.add_buffer buf buf';
    (* '-' must be the first or the last character; '^' and '!' must
     * not be the first character. So we usually print these
     * characters in the order "^!-". One case is special: We have
     * not yet printed any character. Then, "-" must be printed
     * first (if member of the set), or we have one of the very
     * special cases already tested above.
     *)
    if !empty then (
      if want_minus && not done_minus then Buffer.add_char buf '-';
      if want_circum && not done_circum then Buffer.add_char buf '^';
      if want_exclam && not done_exclam then Buffer.add_char buf '!';
    ) else (
      if want_circum && not done_circum then Buffer.add_char buf '^';
      if want_exclam && not done_exclam then Buffer.add_char buf '!';
      if want_minus && not done_minus then Buffer.add_char buf '-';
    );
    Buffer.add_char buf ']';
  )
;;
  

let print_glob_expr (Glob gl) =
  let buf = Buffer.create 200 in
  let rec print gl =
    match gl with
	G_literal s :: gl' ->
	  Buffer.add_string buf (Str.global_substitute esc_re esc_subst s);
	  print gl'
      | G_star :: gl' ->
	  Buffer.add_string buf "*";
	  print gl'
      | G_qmark :: gl' ->
	  Buffer.add_string buf "?";
	  print gl'
      | G_bracket (negated,set) :: gl' ->
	  print_set buf '!' negated set;
	  print gl'
      | G_brace ge_list :: gl' ->
	  Buffer.add_string buf "{";
	  let first = ref true in
	  List.iter 
	    (fun (Glob ge) ->
	       if not !first then Buffer.add_string buf ",";
	       print ge;
	    )
	    ge_list;
	  Buffer.add_string buf "}";
	  print gl'
      | [] ->
	  ()
  in
  print gl;
  Buffer.contents buf
;;


let rec product f l1 l2 =
  match l1 with
      [] ->
	[]
    | x1 :: l1' ->
	List.map (fun x2 -> f x1 x2) l2 @ product f l1' l2
;;


let rec expand_braces_1 (Glob ge) =
  match ge with
      [] ->
	[ [] ]
    | G_brace gelist :: ge' ->
	let gelist' = 
	  List.flatten (List.map expand_braces_1 gelist) in
	let ge_alts' = expand_braces_1 (Glob ge') in
	product ( @ ) gelist' ge_alts'

    | any :: ge' ->
	let ge_alts' = expand_braces_1 (Glob ge') in
	List.map (fun ge_alt' -> any :: ge_alt') ge_alts'
;;


let expand_braces ge =
  List.map (fun ge -> Glob ge) (expand_braces_1 ge)
;;


(* Glob expressions to regular expressions:
 *
 * The translation scheme:
 *
 * (1) = not ~protect_period, not ~protect_slash
 * (2) = not ~protect_period, ~protect_slash
 * (3) = ~protect_period, not ~protect_slash
 * (4) = ~protect_period, ~protect_slash
 *
 *                  (1)            (2)          (3)
 * ========================================================================
 * "*"              .*             [^/]*        [^.]* (at beginning)
 *                                              .* (other positions)
 * "?"              .              [^/]         [^.] (at beginning)
 *                                              . (other positions)
 * "[set]"          [set]          [set']       [set'] (at beginning)
 *                                 (/ deleted)  [set] (other positions)
 *
 *                  (4)
 * ========================================================================
 * "*"              [^/.]* (at beginning, or after a literal "/")
 *                  [^/]* (other positions)
 * "?"              [^/.] (at beginning, or after a literal "/")
 *                  [^/] (other positions)
 * "[set]"          [set'] (/ and . deleted; at beginning or after a lit "/")
 *                  [set''] (/ deleted; other positions)
 *
 * Further rules:
 * - "**" is reduced to "*"
 * - "*?" at the begining (or after /): 
 *   For (3), this matches the same as "?*", i.e. all strings with 
 *   length >= 1 that do not begin with "."
 *   For (4), similar.
 * - "*[set]" at the begining (or after /): 
 *   The case that "*" matches the empty string is made explicit:
 *   ([set']|[^.]+[set])
 *   
 *)


let add_set code set =
  if mem_set code set then
    set
  else
    ( (code,code) :: set )
;;


let rec del_set code set =
  match set with
      [] -> []
    | (x,y) :: set' ->
	if code = x && code = y then
	  del_set code set'
	else if code = x && y > x then
	  (x+1,y) :: del_set code set'
	else if code = y && y > x then
	  (x,y-1) :: del_set code set'
	else if code > x && code < y && y > x then
	  (x,code-1) :: (code+1,y) :: del_set code set'
	else
	  (x,y) :: del_set code set'
;;


let regexp_of_glob_expr ?(protect_period=true) ?(protect_slash=true) (Glob ge) =

  let double_protection last_atom =
    protect_period &&
    match last_atom with
	None ->
	  (* At the very beginning *)
	  true
      | Some(G_literal s) ->
	  (* After a '/' *)
	  protect_slash && s <> "" && s.[ String.length s - 1 ] = '/'
      | _ ->
	  false
  in

  let add_anychar_re buf last_atom =
    match (protect_period, protect_slash) with
	(false, false) ->
	  Buffer.add_string buf "[\000-\255]"
      | (false, true) ->
	  Buffer.add_string buf "[^/]"
      | (true, false) ->
	  if last_atom = None then
	    Buffer.add_string buf "[^.]"
	  else
	    Buffer.add_string buf "[\000-\255]"
      | (true, true) ->
	  if double_protection last_atom then
	    Buffer.add_string buf "[^/.]"
	  else
	    Buffer.add_string buf "[^/]"
  in

  let add_set_re buf last_atom negated set =
    let remove ch set =
      if negated then 
	add_set (Char.code ch) set
      else 
	del_set (Char.code ch) set 
    in
    let set_to_print =
      match (protect_period, protect_slash) with
	  (false, false) ->
	    set
	| (false, true) ->
	    remove '/' set
	| (true, false) ->
	    if last_atom = None then
	      remove '.' set
	    else
	      set
	| (true, true) ->
	    let set' = remove '/' set in
	    if double_protection last_atom then
	      remove '.' set'
	    else
	      set'
    in
    print_set buf '^' negated set_to_print
  in
  
  let rec loop buf last_atom ge =
    match ge with
	(G_literal s as this_atom) :: ge' ->
	  Buffer.add_string buf (Str.quote s);
	  loop buf (Some this_atom) ge'
      | G_star :: ge' ->
	  if last_atom = Some G_star then    (* Adjacent G_stars *)
	    loop buf last_atom ge'
	  else 
	    if protect_period && double_protection last_atom then (
	      (* Special case: Split into the cases that "*" matches
	       * a non-empty string, and that "*" matches the empty string.
	       *)
	      Buffer.add_string buf "\\(";
	      (* Case non-empty string *)
	      add_anychar_re buf last_atom;
	      add_anychar_re buf (Some G_star);
	      Buffer.add_char buf '*';
	      loop buf (Some G_star) ge';
	      Buffer.add_string buf "\\|";
	      (* Case empty string: propagate last_atom to the next
	       * loop iteration
	       *)
	      loop buf last_atom ge';
	      Buffer.add_string buf "\\)";
	    )
	    else (
	      add_anychar_re buf last_atom;
	      Buffer.add_char buf '*';
	      loop buf (Some G_star) ge'
	  )
      | G_qmark :: ge' ->
	  add_anychar_re buf last_atom;
	  loop buf (Some G_qmark) ge'
      | (G_bracket (negated, set) as this_atom) :: ge' ->
	  add_set_re buf last_atom negated set;
	  loop buf (Some this_atom) ge'
      | G_brace _ :: _ ->
	  failwith "Godi_match.regexp_of_glob_expr: brace expression found"
      | [] ->
	  ()
  in
  let buf = Buffer.create 200 in
  Buffer.add_string buf "^";
  loop buf None ge;
  Buffer.add_string buf "$";
  Buffer.contents buf
;;


let slash_re = Str.regexp "/+";;


let split_glob_expr (Glob ge) =

  let rec split_loop is_first acc ge =
    (* acc: accumulates the current component *)
    match ge with
	[] ->
	  [ Glob (List.rev acc) ]
      | (G_literal s as atom) :: ge' ->
	  assert(s <> "");
	  ( try 
	      let k = Str.search_forward slash_re s 0 in (* or Not_found *)
	      let j = Str.match_end() in
	      let l = String.length s in
	      let s1 = String.sub s 0 k in         (* part before '/' *)
	      let s2 = String.sub s j (l - j) in   (* part after '/' *)
	      if is_first && k = 0 then (
		(* Case: rooted expression *)
		let ge'' =
		  if s2 <> "" then (G_literal s2) :: ge' else ge' in
		let comps = split_loop false [] ge'' in
		match comps with
		    Glob ( (G_literal s3) :: r ) :: l ->
		      (Glob (G_literal("/" ^ s3) :: r)) :: l
		  | Glob r :: l ->
		      (Glob (G_literal "/" :: r) :: l)
		  | [] ->
		      [ Glob [ G_literal "/" ] ]
	      )
	      else
		if ge' = [] && s2 = "" then (
		  (* Case: component matches only directory *)
		  [ Glob (List.rev (G_literal (s1 ^ "/") :: acc)) ]
		)
		else (
		  let acc' = 
		    if s1 <> "" then (G_literal s1)::acc else acc in
		  let ge'' =
		    if s2 <> "" then (G_literal s2) :: ge' else ge' in
		  Glob (List.rev acc') :: split_loop false [] ge''
		)
	      
	    with
		Not_found ->
		  split_loop false (atom::acc) ge'
	  )
	    
      | (G_star | G_qmark | G_bracket(_,_) as atom) :: ge' ->
	  split_loop false (atom::acc) ge'

      | G_brace _ :: _ ->
	  failwith "Godi_match.split_glob_expr: brace expression found"
  in

  split_loop true [] ge
;;


let check_rooted_glob_expr (Glob ge) =
  match ge with
      (G_literal s) :: r ->
	if Str.string_match slash_re s 0 then (
	  let j = Str.match_end() in
	  let l = String.length s in
	  let s' = String.sub s j (l - j) in   (* part after '/' *)
	  if s' = "" then 
	    Some (Glob r)
	  else
	    Some (Glob (G_literal s' :: r))
	)
	else
	  None
    | _ ->
	None
;;


let check_directory_glob_expr (Glob ge) =
  match List.rev ge with
      (G_literal s) :: r ->
	( try
	    let l = String.length s in
	    let k = Str.search_backward slash_re s l in  (* or Not_found *)
	    if Str.match_end() < l then raise Not_found;
	    let s' = String.sub s 0 k in   (* the part before '/' *)
	    if s' = "" then
	      Some(Glob(List.rev r))
	    else
	      Some(Glob(List.rev (G_literal s' :: r)))
	  with
	      Not_found -> None
	)
    | _ ->
	None
;;


type fsys =
    { fsys_read_dir : string -> string list;
      fsys_file_is_dir : string -> bool;
      fsys_file_exists : string -> bool;
    }
;;


let fsys_default =
  let read_dir d =
    (* Sys.readdir does not return ".", and "..", so we do not use it *)
    try
      Godi_file.read_dir ~include_dots:true (Godi_file.filename d)
    with
	error ->   (* Cannot open directory *)
	  []
  in
  let file_is_dir f =
    Godi_file.file_is_dir (Godi_file.filename f) in
  let file_exists f =
    Godi_file.file_exists (Godi_file.filename f) in
  { fsys_read_dir = read_dir;
    fsys_file_is_dir = file_is_dir;
    fsys_file_exists = file_exists;
  }
;;

type glob_mode = [ `Existing_paths
		 | `All_paths
		 | `All_words
		 ] ;;

let fn_concat d f =
  let l = String.length d in
  if l = 0 || d.[l-1] = '/' then
    d ^ f
  else
    d ^ "/" ^ f



let glob ?base_dir 
	 ?(protect_period=true) 
	 ?(fsys = fsys_default)
	 ?(mode = `Existing_paths)
	 ge =

  let rec collect_and_match base_dir generated_prefix components =
    match components with
	[] ->
	  if generated_prefix <> "" then [ generated_prefix ] else []
      | comp :: components' ->
	  let full_path file =
	    match base_dir with
		Some d -> fn_concat d file
	      | None   -> file 
	  in

	  let dir_ge = check_directory_glob_expr comp in
	  let comp' = 
	    match dir_ge with 
		Some ge' -> ge'
	      | None -> comp 
	  in

	  let check_for_match only_dirs re file =
	    Str.string_match re file 0 &&
	    (not only_dirs || fsys.fsys_file_is_dir (full_path file))
	  in
	  
	  let files =
	    match comp' with
		Glob [ G_literal s ] ->
		  ( match mode with
			`Existing_paths ->
			  let path = full_path s in
			  if fsys.fsys_file_exists path then
			    [ s ]
			  else
			    []
		      | _ ->
			  [ s ]
		  )
	      | _ ->
		  let only_dirs = components' <> [] || dir_ge <> None in
		  let re = Str.regexp
			     (regexp_of_glob_expr ~protect_period comp') in
		  let file_list = fsys.fsys_read_dir (full_path ".") in
		  List.filter (check_for_match only_dirs re) file_list 
	  in

	  List.flatten
	    (List.map
	       (fun file -> 
		  let prefixed_file = 
		    fn_concat generated_prefix file 
		    ^ (if dir_ge <> None then "/" else "") in
		  
		  collect_and_match 
		    (Some(full_path file))
		    prefixed_file
		    components'
	       )
	       files
	    )

  in

  let collect_and_match_0 components =
    match components with
	comp :: components' ->
	  ( match check_rooted_glob_expr comp with
		None ->
		  collect_and_match base_dir "" components
	      | Some comp' ->
		  if comp' = Glob [] then
		    (* Special case "/" *)
		    [ "/" ]
		  else
		    collect_and_match (Some "/") "/" (comp' :: components')
	  )
      | [] ->
	  []
  in

  let ge_list = expand_braces ge in
  List.flatten
    (List.map
       (fun ge' ->
	  let l = collect_and_match_0 (split_glob_expr ge') in
	  if mode = `All_words && l = [] && ge' <> Glob [] then
	    [print_glob_expr ge']
	  else
	    l
       )
       ge_list
    )
;;


(**********************************************************************)
(* (EXTENDED) REGULAR EXPRESSIONS                                     *)
(**********************************************************************)

(* FIXME: Str does not like \\(^|X\\), i.e. the match of the beginning 
   within an alternative
 *)

let any_regexp flavor s =
  let rec collect_until lexbuf =
    let tok = 
      match flavor with
	| `Extended -> ext_regex_expr lexbuf
	| `Basic    -> basic_regex_expr lexbuf
    in
    if tok = RE_end then
      []
    else
      tok :: (collect_until lexbuf)
  in

  let rec postprocess_basic_re is_first toks =
    (* ^ and $ are only recognized in special places *)
    match toks with
      | RE_lanchor :: toks' when is_first ->
	  RE_lanchor :: postprocess_basic_re false toks'
      | RE_lanchor :: toks' ->
	  RE_char '^' :: postprocess_basic_re false toks'
      | RE_lparen :: RE_lanchor :: toks' ->
	  RE_lparen :: RE_lanchor :: postprocess_basic_re false toks'
      | RE_ranchor :: RE_rparen :: toks' ->
	  RE_ranchor :: RE_rparen :: postprocess_basic_re false toks'
      | [ RE_ranchor ] ->
	  [ RE_ranchor ]
      | RE_ranchor :: toks' ->
	  RE_char '$' :: postprocess_basic_re false toks'
      | other :: toks' ->
	  other :: postprocess_basic_re false toks'
      | [] ->
	  []
  in

  let bracket_to_str buf bl =
    (* Bracket syntax seems to be exactly the same! *)
    List.iter
      (fun b ->
	 match b with
	     Bracket_char c        -> Buffer.add_char buf c
	   | Bracket_range (c1,c2) -> Buffer.add_char buf c1;
	                              Buffer.add_char buf '-';
				      Buffer.add_char buf c2
	   | Bracket_end           -> ()
      )
      bl
  in

  let regexp_to_str buf re =
    List.iter
      (fun re_tok ->
	 match re_tok with
	     RE_alt     -> Buffer.add_string buf "|"
	   | RE_many    -> Buffer.add_string buf "*"
	   | RE_many1   -> Buffer.add_string buf "+"
	   | RE_opt     -> Buffer.add_string buf "?"
	   | RE_lparen  -> Buffer.add_string buf "\\("
	   | RE_rparen  -> Buffer.add_string buf "\\)"
	   | RE_any     -> Buffer.add_string buf "."
	   | RE_lanchor -> Buffer.add_string buf "^"
	   | RE_ranchor -> Buffer.add_string buf "$"
	   | RE_backref c -> 
 	                   Buffer.add_string buf ("\\" ^ String.make 1 c)
	   | RE_char c  -> Buffer.add_string buf (Str.quote (String.make 1 c))
	   | RE_end     -> assert false
	   | RE_bound(n_min, n_max) -> 
	       (* Bounds are not supported by Str, so we can only catch
		* easy cases
		*)
	       ( match (n_min, n_max) with
		     (0, None)   -> Buffer.add_string buf "*"
		   | (1, None)   -> Buffer.add_string buf "+"
		   | (0, Some 1) -> Buffer.add_string buf "?"
		   | (1, Some 1) -> ()
		   | _           -> 
		       raise(Unsupported_expr s)
	       )
	   | RE_bracket (neg, blist) ->
	       Buffer.add_string buf "[";
	       if neg then Buffer.add_string buf "^";
	       bracket_to_str buf blist;
	       Buffer.add_string buf "]";
      )
      re
  in
  
  try
    let re = collect_until (Lexing.from_string s) in
    let re = if flavor = `Basic then postprocess_basic_re true re else re in
    let buf = Buffer.create 80 in
    regexp_to_str buf re;
    (* prerr_endline("DEBUG: " ^ Buffer.contents buf); *)
    Str.regexp (Buffer.contents buf)
  with
      Bracket_Unsupported ->
	raise(Unsupported_expr s)
    | Lexing_Error ->
	raise(Bad_regular_expr s)
    | Failure _ ->
	raise(Bad_regular_expr s)
;;


let basic_regexp = any_regexp `Basic
let extended_regexp = any_regexp `Extended


(**********************************************************************)
(* STRING SUBSTITUTIONS                                               *)
(**********************************************************************)


type subst_expr = (Str.regexp * string) ;;

exception No_percent_symbol

exception Bad_subst_expr


let bs_quote_re = Str.regexp "[\\]";;


let quote_escape s =
  (* Prepend a backslash to every backslash *)
  Str.global_substitute 
    bs_quote_re 
    (fun s -> "\\" ^ Str.matched_string s)
    s
;;


let explode enable_escape s =
  if enable_escape then
    generic_lex (Lexing.from_string s)
  else
    generic_lex_noescape (Lexing.from_string s)
;;


type cstring_atom =
    Literal of string    (* The substring means itself *)
  | Meta of char         (* A meta character has been found *)
;;


let meta_split meta_list s =
  let mkstring acc acclen =
    let u = String.create acclen in
    let i = ref (acclen-1) in
    List.iter
      (fun char -> u.[ !i ] <- char; decr i)
      acc;
    u
  in

  let rec loop acc acclen s =
    match s with
	[] ->
	  if acclen=0 then
	    []
	  else
	    [ Literal (mkstring acc acclen) ]
      | C c :: s' ->
	  if List.mem c meta_list then (
	    if acclen = 0 then
	      Meta c :: loop [] 0 s'
	    else
	      Literal (mkstring acc acclen) :: Meta c :: loop [] 0 s'
	  )
	  else
	    loop (c :: acc) (acclen+1) s'
      | E c :: s' ->
	  loop (c :: acc) (acclen+1) s'
  in
  loop [] 0 s
;;


let parse_percent_subst_expr 
      ?(indicate_no_percent = false)
      ?(enable_escape = true)
      (lhs, rhs) =

  let g_found_percent = ref false in

  let rec rewrite_lhs found_percent l =
    match l with
	[] ->
	  if indicate_no_percent && not found_percent then
	    raise No_percent_symbol;
	  g_found_percent := found_percent;
	  []
      | Meta '%' :: l' ->
	  if found_percent then raise Bad_subst_expr;
	  "\\([\000-\255]*\\)" :: rewrite_lhs true l'
      | Literal s :: l' ->
	  (Str.quote s) :: rewrite_lhs found_percent l'
      | Meta _ :: _ ->
	  assert false
  in

  let rec rewrite_rhs l =
    match l with
	[] ->
	  []
      | Meta '%' :: l' ->
	  if not !g_found_percent then raise Bad_subst_expr;
	  "\\1" :: rewrite_rhs l'
      | Literal s :: l' ->
	  (quote_escape s) :: rewrite_rhs l'
      | Meta _ :: _ ->
	  assert false
  in

  let lhs_cstring = meta_split ['%'] (explode enable_escape lhs) in
  let rhs_cstring = meta_split ['%'] (explode enable_escape rhs) in

  let lhs_re_str = String.concat "" (rewrite_lhs false lhs_cstring) in
  let rhs_repl   = String.concat "" (rewrite_rhs rhs_cstring) in

  (Str.regexp ("^" ^ lhs_re_str ^ "$"), rhs_repl)
;;


let parse_suffix_subst_expr 
      ?(enable_escape = true)
      (lhs, rhs) =

  let rec implode l =
    match l with
	[] -> 
	  []
      | Literal s :: l' ->
	  s :: implode l'
      | _ :: _ ->
	  assert false in

  let lhs_cstring = meta_split [] (explode enable_escape lhs) in
  let rhs_cstring = meta_split [] (explode enable_escape rhs) in

  let lhs' = String.concat "" (implode lhs_cstring) in
  let rhs' = String.concat "" (implode rhs_cstring) in

  let lhs_re_str = "\\([\000-\255]*\\)" ^ Str.quote lhs' in
  let rhs_repl   = "\\1" ^ quote_escape rhs' in

  (Str.regexp ("^" ^ lhs_re_str ^ "$"), rhs_repl)
;;


let parse_substring_subst_expr 
      ?(enable_anchors = true)
      ?(enable_escape = true)
      (lhs, rhs) =

  let rec rewrite_lhs at_beginning l =
    match l with
	[] ->
	  []
      | Meta '^' :: l' ->
	  if at_beginning then
	    "^" :: rewrite_lhs false l'
	  else
	    "\\^" :: rewrite_lhs false l'
      | Meta '$' :: l' ->
	  if l' = [] then
	    "$" :: rewrite_lhs false l'
	  else
	    "\\$" :: rewrite_lhs false l'
      | Literal s :: l' ->
	  (Str.quote s) :: rewrite_lhs false l'
      | Meta _ :: _ ->
	  assert false
  in

  let rec rewrite_rhs l =
    match l with
	[] ->
	  []
      | Meta '&' :: l' ->
	  "\\1" :: rewrite_rhs l'
      | Literal s :: l' ->
	  (quote_escape s) :: rewrite_rhs l'
      | Meta _ :: _ ->
	  assert false
  in

  let lhs_meta_chars =
    if enable_anchors then ['^'; '$'] else [] in

  let lhs_cstring = meta_split lhs_meta_chars (explode enable_escape lhs) in
  let rhs_cstring = meta_split ['&']          (explode enable_escape rhs) in

  let lhs_re_str = String.concat "" (rewrite_lhs true lhs_cstring) in
  let rhs_repl   = String.concat "" (rewrite_rhs rhs_cstring) in

  (Str.regexp ("\\(" ^ lhs_re_str ^ "\\)"), rhs_repl)
;;


let split_ed_subst_expr s =
  let l = String.length s in
  if l < 3 then raise Bad_subst_expr;
  let delim = s.[0] in
  let j = 
    try String.index_from s 1 delim 
    with Not_found -> raise Bad_subst_expr in
  if j = l-1 then raise Bad_subst_expr;
  let k =
    try String.index_from s (j+1) delim
    with Not_found -> raise Bad_subst_expr in
  (String.sub s 1 (j-1),
   String.sub s (j+1) (k-j-1),
   String.sub s (k+1) (l-k-1)
  )


let parse_ed_subst_expr (lhs,rhs) =
  let re = basic_regexp lhs in
  let rl = explode true rhs in
  
  let b = Buffer.create 80 in
  List.iter
    (function
       | E ('1'..'9' as digit) ->
	   Buffer.add_char b '\\';
	   Buffer.add_char b digit
       | E '&' ->
	   Buffer.add_char b '&'
       | E '\\' ->
	   Buffer.add_string b "\\\\"
       | E c ->
	   Buffer.add_char b c
       | C '&' ->
	   Buffer.add_string b "\\0"
       | C c ->
	   Buffer.add_char b c
    )
    rl;

  (re, Buffer.contents b)



let create_subst_expr re repl = (re,repl);;


let substitute2 ?(globally=false) (re,repl) s =
  let have_subst = ref false in
  let subst = 
    if globally then
      Str.global_substitute 
    else
      Str.substitute_first in
  let u =
    subst
      re
      (fun s -> have_subst := true; Str.replace_matched repl s)
      s in
  (!have_subst, u)
;;


let substitute ?globally (re,repl) s =
  snd (substitute2 ?globally (re,repl) s)

(**********************************************************************)
(* UNQUOTING SHELL LISTS                                              *)
(**********************************************************************)

let parse_shell_words s =
  Godi_match_lex.parse_shell_words (Lexing.from_string s)
;;

let sh_quote_re =
  Str.regexp "[^A-Za-z0-9_.:@/-]";;


let quote_shell_word s =
  Str.global_substitute
    sh_quote_re
    (fun s ->
       let m = Str.matched_string s in
       if m = "\r" || m = "\n" then
         "'" ^ m ^ "'"
       else
         "\\" ^ m)
    s

let quote_shell_words l =
  String.concat " " (List.map quote_shell_word l)
